/* Copyright (C) 1995, Sun Microsystems, Inc. */

/*
 * FUNCTION
 *      vdk_vis_fir_16      - finite impulse response filter
 *      vdk_vis_fir_16_32   - 32-point finite impulse response filter
 *      vdk_vis_fir_16_32_a - 32-point finite impulse response filter
 *
 * SYNOPSIS
 *      void vdk_vis_fir_16 (vis_s16 *src, vis_s16 *dst, int dlen, 
 *                           vis_s16 *fir, int flen);
 *      void vdk_vis_fir_16_32 (vis_s16 *src, vis_s16 *dst, int dlen, 
 *                              vis_s16 *fir);
 *      void vdk_vis_fir_16_32_a (vis_s16 *src, vis_s16 *dst, int dlen, 
 *                                vis_s16 *fir);
 *
 * ARGUMENT
 *      src     printer to first sample of source data
 *      dst     printer to first sample of destination data
 *      dlen    length of destination data
 *      fir     coefficients of FIR filter
 *      flen    length of FIR filter
 *
 * DESCRIPTION
 *              flen-1
 *      dst[n] = SUM{fir[k]*src[n+k]}, 0 <= n < dlen
 *               k=0
 */

#include <stdlib.h>
#include "vis_types.h"
#include "vis_proto.h"

/***************************************************************/

#define FIR(ss, ff)                                             \
        sp = (vis_d64 *) vis_alignaddr(ss, off);                \
        s0 = sp[0];                                             \
        s1 = sp[1];                                             \
        sd = vis_faligndata(s0, s1);                            \
        sh = vis_read_hi(sd);                                   \
        sl = vis_read_lo(sd);                                   \
        thh = vis_fmuld8sux16(sh, ff);                          \
        tlh = vis_fmuld8sux16(sl, ff);                          \
        thl = vis_fmuld8ulx16(sh, ff);                          \
        tll = vis_fmuld8ulx16(sl, ff);                          \
        tdh = vis_fpadd32(thh, thl);                            \
        tdl = vis_fpadd32(tlh, tll);                            \
        rdh = vis_fpadd32(rdh, tdh);                            \
        rdl = vis_fpadd32(rdl, tdl);

/***************************************************************/

void vdk_vis_fir_16 (vis_s16 *src, vis_s16 *dst, int dlen, 
                     vis_s16 *fir, int flen)
{
  vis_u8  *sa, *ss;     /* start point in source data */
  vis_d64 *sp;          /* 8-byte aligned start point in source data */
  vis_u8  *da, *dend;   /* start and end points of a line in destination */
  vis_d64 *dp;          /* 8-byte aligned start point in destination */
  int     off;          /* offset of address alignment in destination */
  int     emask;        /* edge masks */
  vis_d64 sd, s0, s1;                   /* source data */
  vis_f32 sh, sl;
  vis_f32 ff;                           /* filter data */
  vis_u32 fu;
  vis_d64 thh, thl, tlh, tll, tdh, tdl; /* termporaries */
  vis_d64 rdh, rdl;                     /* intermediate results */
  vis_d64 dd;                           /* destination data */
  vis_f32 dh, dl;
  int     n, k, num;                    /* loop variables */

  /* set GSR scale factor to 0, so bit 16 to bit 31 of      */
  /* each vis_s32 component will be saved in vis_fpackfix() */
  vis_write_gsr(0);

  /* prepare the detination address */
  da = (vis_u8 *) dst;
  dp = (vis_d64 *) ((vis_u32) da & (~7));
  off = (vis_u32) dp - (vis_u32) da;
  dend = da + 2 * dlen - 1;

  /* generate edge mask for the start point */
  emask = vis_edge16(da, dend);

  /* prepare the source address */
  sa = (vis_u8 *) src;

  num = ((vis_u32) dend >> 3) - ((vis_u32) da >> 3) + 1;

  for (n = 0; n < num; n ++) {

    ss = sa;

    rdh = vis_fzero();
    rdl = vis_fzero();

    for (k = 0; k < flen; k ++) {

      /* load 8 bytes of source data */
      sp = (vis_d64 *) vis_alignaddr(ss, off);
      s0 = sp[0];
      s1 = sp[1];
      sd = vis_faligndata(s0, s1);

      fu = (fir[k] << 16) | (fir[k] & 0xffff);
      ff = vis_to_float(fu);

      sh = vis_read_hi(sd);
      sl = vis_read_lo(sd);

      thh = vis_fmuld8sux16(sh, ff);
      tlh = vis_fmuld8sux16(sl, ff);

      thl = vis_fmuld8ulx16(sh, ff);
      tll = vis_fmuld8ulx16(sl, ff);

      tdh = vis_fpadd32(thh, thl);
      tdl = vis_fpadd32(tlh, tll);

      rdh = vis_fpadd32(rdh, tdh);
      rdl = vis_fpadd32(rdl, tdl);

      ss += 2;
    }

    dh = vis_fpackfix(rdh);
    dl = vis_fpackfix(rdl);
    dd = vis_freg_pair(dh, dl);

    /* store 8 bytes of result */
    vis_pst_16(dd, dp, emask);

    sa += 8;
    dp ++;

    /* prepare edge mask for the end point */
    emask = vis_edge16(dp, dend);
  }
}

/***************************************************************/

void vdk_vis_fir_16_32 (vis_s16 *src, vis_s16 *dst, int dlen, vis_s16 *fir)

{
  vis_u8  *sa, *ss;     /* start point in source data */
  vis_d64 *sp;          /* 8-byte aligned start point in source data */
  vis_u8  *da, *dend;   /* start and end points of a line in destination */
  vis_d64 *dp;          /* 8-byte aligned start point in destination */
  int     off;          /* offset of address alignment in destination */
  int     emask;        /* edge masks */
  vis_d64 sd, s0, s1;                   /* source data */
  vis_f32 sh, sl;
  vis_f32 ff[32];                       /* filter data */
  vis_u32 fu;
  vis_d64 thh, thl, tlh, tll, tdh, tdl; /* termporaries */
  vis_d64 rdh, rdl;                     /* intermediate results */
  vis_d64 dd;                           /* destination data */
  vis_f32 dh, dl;
  int     n, k, num;                    /* loop variables */

  for (k = 0; k < 32; k ++) {
    fu = (fir[k] << 16) | (fir[k] & 0xffff);
    ff[k] = vis_to_float(fu);
  }

  /* set GSR scale factor to 0, so bit 16 to bit 31 of      */
  /* each vis_s32 component will be saved in vis_fpackfix() */
  vis_write_gsr(0);

  /* prepare the detination address */
  da = (vis_u8 *) dst;
  dp = (vis_d64 *) ((vis_u32) da & (~7));
  off = (vis_u32) dp - (vis_u32) da;
  dend = da + 2 * dlen - 1;

  /* generate edge mask for the start point */
  emask = vis_edge16(da, dend);

  /* prepare the source address */
  sa = (vis_u8 *) src;

  num = ((vis_u32) dend >> 3) - ((vis_u32) da >> 3) + 1;

  for (n = 0; n < num; n ++) {

    ss = sa;

    rdh = vis_fzero();
    rdl = vis_fzero();

    for (k = 0; k < 32; k ++) {
      FIR(ss, ff[k]);
      ss += 2;
    }

    dh = vis_fpackfix(rdh);
    dl = vis_fpackfix(rdl);
    dd = vis_freg_pair(dh, dl);

    /* store 8 bytes of result */
    vis_pst_16(dd, dp, emask);

    sa += 8;
    dp ++;

    /* prepare edge mask for the end point */
    emask = vis_edge16(dp, dend);
  }
}

/***************************************************************/

void vdk_vis_fir_16_32_a (vis_s16 *src, vis_s16 *dst, int dlen, vis_s16 *fir)

{
  vis_u8  *sa, *ss;     /* start point in source data */
  vis_d64 *sp;          /* 8-byte aligned start point in source data */
  vis_u8  *da, *dend;   /* start and end points of a line in destination */
  vis_d64 *dp;          /* 8-byte aligned start point in destination */
  int     off;          /* offset of address alignment in destination */
  int     emask;        /* edge masks */
  vis_d64 sd, s0, s1;                   /* source data */
  vis_f32 sh, sl;
  vis_u32 fu;                           /* filter data */
  vis_f32 ff00, ff01, ff02, ff03, ff04, ff05, ff06, ff07;
  vis_f32 ff08, ff09, ff10, ff11, ff12, ff13, ff14, ff15;
  vis_f32 ff16, ff17, ff18, ff19, ff20, ff21, ff22, ff23;
  vis_f32 ff24, ff25, ff26, ff27, ff28, ff29, ff30, ff31;
  vis_d64 thh, thl, tlh, tll, tdh, tdl; /* termporaries */
  vis_d64 rdh, rdl;                     /* intermediate results */
  vis_d64 dd;                           /* destination data */
  vis_f32 dh, dl;
  int     n, num;                    /* loop variables */

  fu = (fir[0] << 16) | (fir[0] & 0xffff);
  ff00 = vis_to_float(fu);
  fu = (fir[1] << 16) | (fir[1] & 0xffff);
  ff01 = vis_to_float(fu);
  fu = (fir[2] << 16) | (fir[2] & 0xffff);
  ff02 = vis_to_float(fu);
  fu = (fir[3] << 16) | (fir[3] & 0xffff);
  ff03 = vis_to_float(fu);
  fu = (fir[4] << 16) | (fir[4] & 0xffff);
  ff04 = vis_to_float(fu);
  fu = (fir[5] << 16) | (fir[5] & 0xffff);
  ff05 = vis_to_float(fu);
  fu = (fir[6] << 16) | (fir[6] & 0xffff);
  ff06 = vis_to_float(fu);
  fu = (fir[7] << 16) | (fir[7] & 0xffff);
  ff07 = vis_to_float(fu);
  fu = (fir[8] << 16) | (fir[8] & 0xffff);
  ff08 = vis_to_float(fu);
  fu = (fir[9] << 16) | (fir[9] & 0xffff);
  ff09 = vis_to_float(fu);
  fu = (fir[10] << 16) | (fir[10] & 0xffff);
  ff10 = vis_to_float(fu);
  fu = (fir[11] << 16) | (fir[11] & 0xffff);
  ff11 = vis_to_float(fu);
  fu = (fir[12] << 16) | (fir[12] & 0xffff);
  ff12 = vis_to_float(fu);
  fu = (fir[13] << 16) | (fir[13] & 0xffff);
  ff13 = vis_to_float(fu);
  fu = (fir[14] << 16) | (fir[14] & 0xffff);
  ff14 = vis_to_float(fu);
  fu = (fir[15] << 16) | (fir[15] & 0xffff);
  ff15 = vis_to_float(fu);
  fu = (fir[16] << 16) | (fir[16] & 0xffff);
  ff16 = vis_to_float(fu);
  fu = (fir[17] << 16) | (fir[17] & 0xffff);
  ff17 = vis_to_float(fu);
  fu = (fir[18] << 16) | (fir[18] & 0xffff);
  ff18 = vis_to_float(fu);
  fu = (fir[19] << 16) | (fir[19] & 0xffff);
  ff19 = vis_to_float(fu);
  fu = (fir[20] << 16) | (fir[20] & 0xffff);
  ff20 = vis_to_float(fu);
  fu = (fir[21] << 16) | (fir[21] & 0xffff);
  ff21 = vis_to_float(fu);
  fu = (fir[22] << 16) | (fir[22] & 0xffff);
  ff22 = vis_to_float(fu);
  fu = (fir[23] << 16) | (fir[23] & 0xffff);
  ff23 = vis_to_float(fu);
  fu = (fir[24] << 16) | (fir[24] & 0xffff);
  ff24 = vis_to_float(fu);
  fu = (fir[25] << 16) | (fir[25] & 0xffff);
  ff25 = vis_to_float(fu);
  fu = (fir[26] << 16) | (fir[26] & 0xffff);
  ff26 = vis_to_float(fu);
  fu = (fir[27] << 16) | (fir[27] & 0xffff);
  ff27 = vis_to_float(fu);
  fu = (fir[28] << 16) | (fir[28] & 0xffff);
  ff28 = vis_to_float(fu);
  fu = (fir[29] << 16) | (fir[29] & 0xffff);
  ff29 = vis_to_float(fu);
  fu = (fir[30] << 16) | (fir[30] & 0xffff);
  ff30 = vis_to_float(fu);
  fu = (fir[31] << 16) | (fir[31] & 0xffff);
  ff31 = vis_to_float(fu);

  /* set GSR scale factor to 0, so bit 16 to bit 31 of      */
  /* each vis_s32 component will be saved in vis_fpackfix() */
  vis_write_gsr(0);

  /* prepare the detination address */
  da = (vis_u8 *) dst;
  dp = (vis_d64 *) ((vis_u32) da & (~7));
  off = (vis_u32) dp - (vis_u32) da;
  dend = da + 2 * dlen - 1;

  /* generate edge mask for the start point */
  emask = vis_edge16(da, dend);

  /* prepare the source address */
  sa = (vis_u8 *) src;

  num = ((vis_u32) dend >> 3) - ((vis_u32) da >> 3) + 1;

  for (n = 0; n < num; n ++) {

    ss = sa;

    rdh = vis_fzero();
    rdl = vis_fzero();

    FIR(ss, ff00);
    ss += 2;
    FIR(ss, ff01);
    ss += 2;
    FIR(ss, ff02);
    ss += 2;
    FIR(ss, ff03);
    ss += 2;
    FIR(ss, ff04);
    ss += 2;
    FIR(ss, ff05);
    ss += 2;
    FIR(ss, ff06);
    ss += 2;
    FIR(ss, ff07);
    ss += 2;
    FIR(ss, ff08);
    ss += 2;
    FIR(ss, ff09);
    ss += 2;
    FIR(ss, ff10);
    ss += 2;
    FIR(ss, ff11);
    ss += 2;
    FIR(ss, ff12);
    ss += 2;
    FIR(ss, ff13);
    ss += 2;
    FIR(ss, ff14);
    ss += 2;
    FIR(ss, ff15);
    ss += 2;
    FIR(ss, ff16);
    ss += 2;
    FIR(ss, ff17);
    ss += 2;
    FIR(ss, ff18);
    ss += 2;
    FIR(ss, ff19);
    ss += 2;
    FIR(ss, ff20);
    ss += 2;
    FIR(ss, ff21);
    ss += 2;
    FIR(ss, ff22);
    ss += 2;
    FIR(ss, ff23);
    ss += 2;
    FIR(ss, ff24);
    ss += 2;
    FIR(ss, ff25);
    ss += 2;
    FIR(ss, ff26);
    ss += 2;
    FIR(ss, ff27);
    ss += 2;
    FIR(ss, ff28);
    ss += 2;
    FIR(ss, ff29);
    ss += 2;
    FIR(ss, ff30);
    ss += 2;
    FIR(ss, ff31);
    ss += 2;

    dh = vis_fpackfix(rdh);
    dl = vis_fpackfix(rdl);
    dd = vis_freg_pair(dh, dl);

    /* store 8 bytes of result */
    vis_pst_16(dd, dp, emask);

    sa += 8;
    dp ++;

    /* prepare edge mask for the end point */
    emask = vis_edge16(dp, dend);
  }
}

/***************************************************************/
